/*
 * anyRemote
 * a bluetooth remote for your PC.
 * Copyright (C) 2006
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */
#define _GNU_SOURCE
#include <stdio.h>

#include <stdlib.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "atsend.h"
#include "cmds.h"
#include "common.h"
#include "parse.h"

extern void closePort(int final);

extern int  	    serverMode;
extern char 	    callerId[MAXLEN];
extern int  	    repeatNow;
extern type_key*    repeatCmdPtr;

static void updateTimerLastData (timerCmd* timer, char *data);

char tmptext[MAXCMDLEN];
int  remoteOn       = 1;

timerCmd* timers  = NULL;

// This is not clean-up function. It just finished the connection.
static void finishAll()
{
	char cmd[32];
        char answer[1024];

        // finish events forwarding
        printf("Finish events forwarding\n");
        logger("INFO","Finish events forwarding");

        if (serverMode == 0) {
        	strcpy(cmd,getAT_CMER(CMER_OFF));

                strcat(cmd,"\r");
                sendSyncCommand(cmd, answer, sizeof(answer), NULL);

                //handle_command(answer); // Does it needed ?

                // and resend No back
                //sendSyncCommand("AT+CKPD=\"e\"\r", answer, sizeof(answer), NULL);    // Does it needed ?
        }
        closePort(1);

        return;
}

static int execInFork(char *cmd)
{
	int ret;
        int cpid = fork();
        if( cpid < 0) {
        	logger("ERROR", "Can not fork!");
                return -1;
        }

        if(cpid) {      // father
        	waitpid(cpid,&ret,0);
                if(ret!=-1 && WIFEXITED(ret)) {
                	ret = WEXITSTATUS(ret);
                }
        } else {        // child
        	logger("INFO", "Close port in child");
                closePort(0);
                 
                ret = system(cmd);
                if(ret!=-1 && WIFEXITED(ret)) {
                	ret = WEXITSTATUS(ret);
                }
                exit(ret);
        }

        return ret;
}

static char* fBuffer;

static void freeResultsBuffer() 
{
	free(fBuffer);
	fBuffer = NULL;
}

static char *readResultsFile() 
{
	FILE *fp;
	struct stat buf;
	
	if(stat(TOFILE, &buf) == -1) {
		return NULL;
	}
	
	long fLen = buf.st_size;
	sprintf(tmp, "readResultsFile >%ld<", fLen);
	logger("INFO",tmp);
	
	fp=fopen(TOFILE,"r");
	if (fp == NULL) {
		return NULL;
	}
    
    	fBuffer = calloc(fLen+1,1);
    
	fread(fBuffer, sizeof(char), fLen, fp);
    
	return fBuffer;
}

static int execSimpleCmd(char *command, int toFile, cmdParams *params)
{
	char cmd [1024];
        cmd[0] = '\0';

        if (command == NULL) {
                logger("ERROR","execSimpleCmd(): null input");
                return EXIT_NOK;
        }
        //sprintf(tmp, "execSimpleCmd >%s<", command);
        //logger("INFO",tmp);

        if (*command == '\0') { // Empty command. Skip it.
            return EXIT_OK;
        }

        if (remoteOn != 1) {
                logger("DEBUG", "Remote if OFF. Skip event");
                return EXIT_OK;
        }
        char *papameterized = NULL;

        int pc = strlen(PARAM_CALLID);
        int pt = strlen(PARAM_TIME);
        int pm = strlen(PARAM_MODE);
        int pp = strlen(PARAM_PARAM);
        int pi = strlen(PARAM_INDEX);
         
        while(command != NULL) {
        	sprintf(tmp, "cur cmd >%s<", command);
                logger("DEBUG",tmp);

                papameterized = strstr(command, PARAM_START);   // $(
        
                if (papameterized != NULL) {
                	sprintf(tmp, "papameterized >%s<", papameterized);
                        logger("DEBUG",tmp);
                }
        
                if (papameterized == NULL) { // Just copy
                	logger("DEBUG", "command is not parametrized");
                        strcat(cmd, command);
                        break;
                } else if(strncmp(papameterized+2,PARAM_CALLID,pc) == 0) {
                	
                        *papameterized = '\0';
                
                        // First part
                        strcat(cmd, command);
                
                        // Add params
                        strcat(cmd, callerId);
                
                        command = papameterized + pc + 2;
                    
                } else if (strncmp(papameterized + 2,PARAM_TIME,pt) == 0) { 
                	
                        *papameterized = '\0';

                        // First part
                        strcat(cmd, command);
                
                        // Add time
                        time_t logtime;
                        time(&logtime);

                        struct tm curtime;
                        localtime_r(&logtime, &curtime);

                        char buf[64];
                        sprintf(buf,"%d:%02d", curtime.tm_hour, curtime.tm_min);

                        strcat(cmd, buf);

                        command = papameterized + pt + 2;
                
                } else if (strncmp(papameterized + 2,PARAM_MODE,pm) == 0) {
                	
                        *papameterized = '\0';

                        // First part
                        strcat(cmd, command);
                
                        // Add params
                        strcat(cmd, getMode());
                                
                        command = papameterized + pm + 2;
                
                } else if (strncmp(papameterized + 2,PARAM_PARAM,pp) == 0) {
                	// This is dirty trick to make Bemused commands like VOLM<volume> works
                        // also it is used in Java Client list handling
                
                        *papameterized = '\0';

                        if (params == NULL) {   // value can be empty
                        	logger("ERROR", "No parameters supplied!");
                                return EXIT_OK;
                        }
                
                        // First part
                        strcat(cmd, command);

                        // Add params
                        strcat(cmd, params->value);
                
                        command = papameterized + pp + 2;

                } else if (strncmp(papameterized + 2,PARAM_INDEX,pi) == 0) {
                	// Used in Java Client list handling
                
                        *papameterized = '\0';

                        if (params == NULL || params->index[0] == '\0') {
                        	logger("ERROR", "No parameters supplied!");
                                return EXIT_OK;
                        }
                
                        // First part
                        strcat(cmd, command);

                        // Add params
                        strcat(cmd, params->index);
                
                        command = papameterized + pi + 2;
                
                } else {    // Just copy
                	strcat(cmd, command);
                        break;
                }
                sprintf(tmp, "partial cmd >%s<", cmd);
                logger("DEBUG",tmp);
        }
        sprintf(tmp, "final cmd >%s<", cmd);
        logger("DEBUG",tmp);

        if (toFile) {       // We needs to wait results, so do not run in background
        	strcat(cmd, " > ");
                strcat(cmd, TOFILE);
        } else {
        	if (! index(cmd, '&')) {
                	strcat(cmd, "&");
                }
        }
        return (execInFork(cmd) == -1 ? EXIT_NOK : EXIT_OK);
}

static int processSendCmd(char *origCmd)
{
	int ret = EXIT_OK;

        char all[MAXMAXLEN];
        strncpy(all, origCmd, MAXMAXLEN-1);

        char *tag  = strtok(all, ",");
        if (tag == NULL) {
        	logger("DEBUG", "processSendCmd(): incorrect input (1)");                
        	return EXIT_NOK;
        }
        if (strlen(origCmd) <= strlen(tag) + 1) {   // Set(tag,) or Set(tag) without any value
        	logger("DEBUG", "processSendCmd(): incorrect input (2)");                
                return EXIT_NOK;
        }
        char *ptr   = tag + strlen(tag) - 1;
        // strip spaces from tail
        while( *ptr == '\0' || *ptr == '\n' || *ptr == '\t' || *ptr == ' ') {
        	*ptr = '\0';
        	ptr--;
        }
        ptr = tag + strlen(tag) + 1;
        // strip spaces from head of value
        while( *ptr != '\0' && (*ptr == '\t' || *ptr == ' ')) {
        	ptr++;
        }

        sprintf(tmp, "processSendCmd() is >%s< >%s<",tag,ptr);
        logger("DEBUG", tmp);

        if (strcmp(tag, SEND_STRING) == 0) {
        	ret = writeSocketConn(ptr);
                if (ret != EXIT_OK) {
                	logger("DEBUG","Fails in writeSocketConn()");
                }
        } else if (strcmp(tag, SEND_BYTES) == 0) {
        	ret = writeBytesSocketConn(ptr);
                if (ret != EXIT_OK) {
                	logger("DEBUG","Fails in writeBytesSocketConn()");
                }    
        } else {
        	logger("DEBUG", "processSendCmd(): incorrect input (3)");                
                return EXIT_NOK;
        }
        return ret;
}

static void sendData(int type, char *tag, char *data) 
{
	int ret;
	if (type == ID_SEND && strcmp(tag,SEND_BYTES) == 0) {
		ret = writeBytesSocketConn(data);
	} else {
		ret = writeSocketConn(data);
	}
	if (ret != EXIT_OK) {
		logger("DEBUG","Fails in write(Bytes)SocketConn()");
	}
}

static int macroCmd(char *macro, cmdParams *params) 
{
 	int ret;
	type_key *k = findItem(macro, &ret, NULL);

	if (k && ret == FLAG_EXACT) {
		ret = handleCmdByKey(k, params);
		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
			return ret;
		}
	}
	return EXIT_OK;
}

static int loadCmds(char *file) 
{
	FILE *fp;
        char *command = NULL;
        size_t len = 0;
        ssize_t read;
	
	fp=fopen(file,"r");
	if (fp == NULL) {
		return EXIT_NOK;
	}
        
	while ((read = getline(&command, &len, fp)) != -1) {
        	sprintf(tmp,"Retrieved line %s of length %d ", command, (int) read);
		logger("DEBUG",tmp);
                
        	cmdItem* tmpCmd = NULL;
                storeCmds(&tmpCmd, command); // parse command on-the-fly 
                
                processCommands(tmpCmd,NULL);
                
                freeCmds(tmpCmd);
        }
        if (command) {
        	free(command);
        }
	return EXIT_OK;
}

static int execCmdAndGetResults(int type, char *commandIn, timerCmd* timer, cmdParams *params) 
{
	logger("DEBUG","execCmdAndGetResults");

        if (serverMode != SERVER_NO && remoteOn == 1 && commandIn != NULL) {
        	
                // Parse type of set
                char commandOrig[MAXMAXLEN];
                strncpy(commandOrig, commandIn, MAXMAXLEN-1);
        
                // commandIn = tag,command or tag,oper,title,command
                char* tag    = strtok(commandOrig,",");
                char* part1  = NULL;
                char* part2  = NULL;
                int commas   = 1;
                int part1len = 0;
                int part2len = 0;
                int taglen     = 0;

                if (tag == NULL) {
                    logger("DEBUG", "execCmdAndGetResults(): incorrect input (1)");                
                    return EXIT_NOK;
                }
        
                taglen = strlen(tag);
        
                if (strstr(tag,SET_TEXT) != NULL || 
                    strstr(tag,SET_LIST) != NULL ||
                    strstr(tag,SET_FMGR) != NULL) { // list|text|filemanager,title,oper,cmd or list|text,fg|bg|font|select,cmd
                        part1 = strtok(NULL,",");
                        if (part1 != NULL) {
                                part1len = strlen(part1);
                                
                                if (strcmp("fg",    part1) == 0 ||
                                    strcmp("bg",    part1) == 0 ||
                                    strcmp("font",  part1) == 0 ||
                                    strcmp("select",part1) == 0) {
                                    commas = 2;
                                }
                        }
                        if (commas != 2) {
                        	part2 = strtok(NULL,",");
                        	if (part2 != NULL) {
                                	part2len = strlen(part2);
                        	}
                        	commas = 3;
                        }
                }
                if (strstr(tag,SET_MENU) != NULL) {
                        part1 = strtok(NULL,",");
                        if (part1 != NULL) {
                                part1len = strlen(part1);
               		}
                        commas = 2;
                }
        
                sprintf(tmp, "execCmdAndGetResults() is >%s< >%s<",tag,commandIn);
                logger("DEBUG", tmp);
        
                if (type == ID_SET && strlen(commandIn) <= taglen + part1len + part2len + commas) {   // Set(tag,) or Set(tag) without any value
                        logger("DEBUG", "execCmdAndGetResults(): incorrect input (2)");                
                        return EXIT_NOK;
                }
                if (type != ID_SET && type != ID_SEND ) {
                        logger("DEBUG", "execCmdAndGetResults(): incorrect input (3)");                
                        return EXIT_NOK;
                }
                char *command2exec = tag + taglen + part1len + part2len + commas;
        
                while(*tag == ' ' || *tag == '\t') {
                        tag++;
                }
        
                int ret = execSimpleCmd(command2exec,1,params);
                if (ret != EXIT_OK) {
                        sprintf(tmp, "Return code from execSimpleCmd() is %d",ret);
                	logger("DEBUG", tmp);
                        return ret;
                }

                char *res  = readResultsFile();
                int resLen = strlen(res);
        
                if (timer != NULL && timer->lastValue != NULL && strcmp(timer->lastValue,res) == 0) {
                        logger("DEBUG","No needs to send the same data.");
                     
                        free(res);
                        return EXIT_OK;
                }
                if (timer != NULL) {    // Update "last data"
                        updateTimerLastData(timer, res);
                }
        
                if (type == ID_SEND) {  // Just send string
                        sendData(type, tag, res);
                } else {                // ID_SET

                        // Replace all ");" inside buffer to "@@" since Java Client will separate commands by ");"
                        char *p;
                        while ((p = strstr(res,");")) != NULL) {
                                *p = '@';
                                p++;
                                *p = '@';
                        }

                        char chunk[256];    
                        strcpy(chunk,CMD_SET);

                        strcat(chunk,"(");  // For example, Set(title,
                        strcat(chunk,tag);
                        strcat(chunk,",");
                    
                        if (strcmp(tag,SET_TEXT) == 0) {
                        	// long commands
                                // Because of bug in Java String class in SE and Nokia (fails to get correct indexOf() 
                                // if string longer than 256 characters) we have to separate result to several chunks
                        
                                int doLoop  = 1;   
                                int headLen = strlen(chunk);     
                                while (doLoop > 0) {
                        
                                        if (part1 != NULL) {
                                                if (strcmp(part1,"replace") == 0) {
                                                        if (doLoop == 1) {
                                                                strcat(chunk,part1);
                                                        } else {
                                                                strcat(chunk,"add");
                                                        }
                                                } else {
                                                        strcat(chunk,part1);
                                                }    
                                        }
                                        if (part2 != NULL) {
                                                strcat(chunk,",");
                                                strcat(chunk,part2);
                                        }
                                        strcat(chunk,",");
                        
                                        int curLen = strlen(chunk);
                                        
                                        if (doLoop >= 18) {     // send only about 4K
                                                strncat(chunk, res, 200 - curLen);
                                                res += (200 - curLen); 
                                                strcat(chunk,"!!! file too long to show it in full !!!");
                                                doLoop = -1;
                                        } else if (strlen(res) + curLen > 251) {       // 253 + ">>);\0" = 256
                                                strncat(chunk, res, 251 - curLen);
                                                res += (251 - curLen);              // and move pointer
                                                strcat(chunk,">>");                 // continuation flag
                                                doLoop++;
                                        } else {                                // last portion
                                                strcat(chunk, res);
                                                doLoop = -1;
                                        }
                        
                                        strcat(chunk,");");
        
                                        //sprintf(tmp, "set-text %d >%s<", doLoop,chunk);
                                        //logger("INFO",tmp);
                                        
                                        sendData(type, tag, chunk);
                                        *(chunk + headLen) = '\0';
                                }
                        
                                logger("INFO","set-text finished");
                       } else if (strcmp(tag,SET_LIST) == 0 || strcmp(tag,SET_FMGR) == 0 || strcmp(tag,SET_MENU) == 0) {
                               // long commands
                               // Because of bug in Java String class in SE and Nokia (fails to get correct indexOf() 
                               // if string longer than 256 characters) we have to separate result to several chunks

                               int doLoop = 1;        
                               int headLen = strlen(chunk);     
                               char* onePart = strtok(res,",");
                               while (doLoop > 0) {
                                       
                                       if (part1 != NULL) {
                                        	if (strcmp(part1,"replace") == 0) {
                                                         if (doLoop == 1) {
                                                          	strcat(chunk,part1);
                                                         } else {
                                                         	strcat(chunk,"add");
                                                         }
                                                } else {
                                                	strcat(chunk,part1);
                                                }    
                                            }
					if (part2 != NULL) {
                				strcat(chunk,",");
                				strcat(chunk,part2);
        				}
                               
					int hEnd = strlen(chunk);
                               
                                      	while (onePart != NULL && strlen(chunk) + strlen(onePart) < 252) {
                                                 strcat(chunk, ",");
                                                 strcat(chunk, onePart);
                                                 onePart = strtok(NULL,",");
                                      	}
                               
                                    	if (strlen(chunk) == hEnd && resLen != 0) { // We fall into infinite loop
                                        	strcat(chunk, ",");
                                                strcat(chunk, "String too long");
                                                onePart = strtok(NULL,",");
                                    	}
                                     	if (resLen == 0) {
                                      		strcat(chunk, ",");
                                     	}
                               
                                     	if (onePart == NULL) {
                                    		doLoop = -1;    // Last portion
                                     	} else {
                                               doLoop++;
                                    	}

                                    	strcat(chunk,");");
                               
                                   	//sprintf(tmp, "set-list|filemanager %d >%s<", doLoop,chunk);
                                	//logger("INFO",tmp);
                               
                                	sendData(type, tag, chunk);
                               		*(chunk + headLen) = '\0';
                       		}
            		} else {    // short commands. Are all of them really short (<256 bytes?)
                               strcat(chunk,res);
                               strcat(chunk,");");
               
                               sendData(type, tag, chunk);
                  	}
                }
                freeResultsBuffer();
        }
        logger("INFO","execCmdAndGetResults finished");
        return EXIT_OK;
}

static int processOneCommand(int cmdType, char *commandIn, cmdParams* p)
{
	 int ret;
         char commandCopy[MAXMAXLEN];
         char *commandOrig = NULL;

         if (commandIn == NULL) {
         	sprintf(tmp, "execCmd >no params< >%d<", remoteOn);
         } else {
                 sprintf(tmp, "execCmd >%s< >%d<", commandIn, remoteOn);
                 strncpy(commandCopy, commandIn, MAXMAXLEN-1);
                 commandOrig = commandCopy;
         }
         logger("INFO",tmp);
         
         
         switch (cmdType) {
         
             case ID_EXIT:               
                 if (remoteOn == 1) {            // User asks to exit        
                         logger("INFO", "Exit key detected");
                         finishAll();                        

                         return EXIT_ABORT;
                 }
                 break;
                 
             case ID_TMPDISCONN:
                 if (remoteOn == 1) {            // User asks to temporary disconnect
                         logger("INFO", "Temporary disconnect key detected");
                         finishAll();
                 
                         return EXIT_DISCON; // But what with the rest of the command line ? (now just skip)
                 }
                 break;
                 
             case ID_REMOTEON:
                 logger("INFO", "Switch remote on");
                 remoteOn = 1;
                 break;
                 
             case ID_REMOTEOFF:
                 logger("INFO", "Switch remote off");
                 remoteOn = 0;
                 break;
                 
             case ID_EXEC:
                 logger("INFO", "Command: Exec");
                 if (remoteOn == 1) {
                 	execSimpleCmd(commandOrig, 0, p);
                 }
                 break;
                 
             case ID_SETMODE:
                 logger("INFO", "Command: SetMode");
                 if (remoteOn == 1) {
                         
                         int i = 0;    
                         type_key* k = findItem(EVT_EXITMODE, &i, NULL);
                         if (k && i == FLAG_EXACT) {
                                 logger("DEBUG", "Exec cmd on exit from mode");
                                 ret = handleCmdByKey(k,NULL);
                                 if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                                 	return ret;
                                 }
                         } else {
                         	logger("DEBUG", "Leave mode");
                         }
                         setMode(commandOrig);
                 
                         i = 0; 
                         k = findItem(EVT_ENTRMODE, &i, NULL);
                         if (k && i == FLAG_EXACT) {
                                 logger("DEBUG", "Exec cmd on enter to mode");
                                 ret = handleCmdByKey(k,NULL);
                                 if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                                 	return ret;
                                 }
                         } else {
                         	logger("DEBUG", "Entered to mode");
                         }
                 } else {
                         logger("DEBUG", "remote Off ???");
                 }
                 break;

             case ID_SENDCKPD:
                 logger("INFO", "Command: SendCKPD");
                 if (serverMode == SERVER_NO && remoteOn == 1) {
                         
                         ret = sendSeq(commandOrig);
                         if (ret != EXIT_OK) {
                         	logger("DEBUG","Fails in sendSeq()");
                         }
                         if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                         	return ret;
                         }
                 }    
                 break;

             case ID_SEND:
                 logger("INFO", "Command: Send");
                 if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                         ret = processSendCmd(commandOrig);
                         if (ret != EXIT_OK) {
                         	logger("DEBUG","Fails in processSendCmd()");
                         }
                 }    
                 break;
                 
                 
             case ID_SET:          
                 logger("INFO", "Command: Set");
                 if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                         strcpy(tmptext, CMD_SET);
                         strcat(tmptext, "(");
                 
                         // Change \n in input to real '\n'
                         char *p1 = commandOrig;
                         char *p2 = commandOrig;
                         while (p2 != NULL) {
                                 p2 = strstr(p1, "\\n");
                         
                                 if (p2 == NULL) {
                                 	strncat(tmptext, p1, p2-p1);
                                 } else {
                                         strncat(tmptext, p1, p2-p1);
                                         strcat(tmptext, "\n");
                                 }
                                 p1 = p2 + 2;
                         }
                         //strcat(tmptext, commandOrig);
                         strcat(tmptext, ");");
                 
                         ret = writeSocketConn(tmptext);
                         if (ret != EXIT_OK) {
                         	logger("DEBUG","Fails in writeSocketConn() (Set)");
                         }
                 }    
                 break;
                 
             case ID_EXECSET:   
                 logger("INFO", "Command: ExecAndSet");
                 
                 execCmdAndGetResults(ID_SET, commandOrig, NULL, p);
                 
                 logger("INFO", "Command: ExecAndSet FINISHED");
                 break;
                 
             case ID_EXECSEND:  
                 logger("INFO", "Command: ExecAndSend");
                 
                 execCmdAndGetResults(ID_SEND, commandOrig, NULL, p);
                 break;
                 
             case ID_TIMERCMD:   
                 logger("INFO", "Command: Timer");
                 if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                 	return processTimerCmd(commandOrig);
                 }     
                 break;

             case ID_CANCELTIMER:   
                 logger("INFO", "Command: CancelTimer");
                 if (serverMode != SERVER_NO && remoteOn == 1 && commandOrig != NULL) {
                 	cancelTimer(commandOrig);
                 }  
                 break;

             case ID_MACRO:   
                 logger("INFO", "Command: Macro");
                 if (remoteOn == 1 && commandOrig != NULL) {
                 	ret = macroCmd(commandOrig, p);
               		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	return ret;
                        }
                 }  
                 break;

             case ID_LOAD:   
                 logger("INFO", "Command: Load");
                 if (remoteOn == 1 && commandOrig != NULL) {
                 	ret = loadCmds(commandOrig);
               		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	return ret;
                        }
                 }  
                 break;
                 
             default:  
                 logger("INFO", "Unknown command");                 
                 break;
        }

        return EXIT_OK;
}

int processCommands(cmdItem* cmdList,cmdParams* p) 
{
	while (cmdList) {
        	if (cmdList->descr == NULL) {
                        sprintf(tmp, "Process next command NULL parameters");
                } else {
                        sprintf(tmp, "Process next command %s",cmdList->descr);
                }
                logger("DEBUG", tmp);
                    
                int ret = processOneCommand(cmdList->type, cmdList->descr, p);
                if (ret != EXIT_OK) {
                	sprintf(tmp, "Return code from processOneCommand() is %d",ret);
                        logger("DEBUG", tmp);
                         
                        if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	logger("DEBUG", "Abort or disconnect after processOneCommand()");
                        
                                return ret;
                        }
                }
                cmdList = cmdList->next;
        }
        return EXIT_OK;
}

//
// Top-level command processing function
//
int handleCmdByKey(type_key* k, cmdParams* p)
{
	if (k == NULL) {
        	logger("DEBUG", "handleCmdByKey() NULL input");
                return EXIT_OK;
        }

        sprintf(tmp, "handleCmdByKey() >%s<",k->key);
        logger("DEBUG", tmp);

        cmdItem* cmdList = getCommand(k);
        if (cmdList == NULL) {
        	logger("DEBUG", "No appropriate command was found by getCommand()");
        } else {
        	sprintf(tmp, "ready to execute command");
                logger("DEBUG", tmp);
        
                // In AT mode: User had pressed some key ... 
                // Send ToMainMenu sequence of CKPD's to show main menu again
                if (serverMode == SERVER_NO) {
                	if (sendSeq(getToMainMenu()) != EXIT_OK) {
                        	logger("DEBUG","Fails in sendSeq()");
                        }
                }
        
                int ret = processCommands(cmdList, p);
                if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                	repeatNow    = 0;
                        repeatCmdPtr = NULL;
                        return ret;
                }
        }   

        return EXIT_OK;     
}

//
// Functions related to timer commands
//
// Timer(id,timeout,maxtimes,tag,command);
//      where tag could be: icons, title, status
//

timerCmd* findTimer(char *id, timerCmd **prev) 
{
	timerCmd* ptr = timers;

        while (ptr != NULL && strncmp(ptr->id,id,MTEXTLEN) != 0) {
        	*prev = ptr;
                //sprintf(tmp, "findTimer(): set prev to >%s<", (*prev)->id);
                //logger("DEBUG",tmp);
                ptr  = ptr->next;
        }

        return ptr;
}

static int createTimer(char* id, int type, float tmout, int times, char *cmd)
{
	if (id == NULL || cmd == NULL) {
        	logger("DEBUG", "createTimer(): null input");                
                return EXIT_NOK;
        }
        timerCmd* ptr = NULL;
        if (findTimer(id, (timerCmd**) &ptr) != NULL) {
        	logger("DEBUG", "createTimer(): timer already exists");                
                return EXIT_NOK;
        } 
        timerCmd *tm = calloc(1,sizeof(timerCmd));

        strncpy(tm->id, id, MTEXTLEN);
        tm->type      = type;
        tm->ticks     = 0;
        tm->times     = 1;
        tm->tmout     = tmout;
        tm->maxTimes  = times;
        tm->lastValue = NULL;

        tm->cmd = (char *) calloc(1, strlen(cmd) + 1);
        strcpy(tm->cmd, cmd);

        // Insert in head
        if (timers == NULL) { // Insert first
        	tm->next = NULL;
        } else {
        	tm->next = timers;
        }
        timers = tm;

        return EXIT_OK;
}

int cancelTimer(char *id) 
{
	sprintf(tmp, "cancelTimer() >%s<", id);
        logger("DEBUG", tmp);

        timerCmd *prev = NULL;
        timerCmd *tm = findTimer(id, (timerCmd **) &prev);
        if (tm == NULL) {
            logger("DEBUG", "Cant find timer to cancel!");
            return EXIT_NOK;
        }    
        if (tm == timers) { //remove in head
        	timers = tm->next;
        } else {
        	if (prev != NULL) {
                	prev->next = tm->next;
                } else {
                	logger("ERROR", "Previous item absent for non-first element");
                        return EXIT_NOK;
                }
        }        

        free(tm->cmd); tm->cmd = NULL;
        free(tm);

        return EXIT_OK;
}


int processTimerCmd(char *origCmd)
{
	char all[MAXMAXLEN];
        strcpy(all, origCmd);

        char *id  = strtok(all, ",");
        if (id == NULL) {
        	logger("DEBUG", "processTimerCmd(): incorrect input - timer data absent");                
                return EXIT_NOK;
        }

        char *ptr = id + strlen(id) - 1;
        // strip spaces from tail
        while( *ptr == '\0' || *ptr == '\n' || *ptr == '\t' || *ptr == ' ') {
        	*ptr = '\0';
                ptr--;
        }

        char* t1  = strtok(NULL, ",");
        char* t2  = strtok(NULL, ",");
        char* cmd = strtok(NULL, "\0");  
        if (t1 == NULL || t2 == NULL || cmd == NULL) {
        	logger("DEBUG", "processTimerCmd(): incorrect input - command is not full");                
                return EXIT_NOK;
        }
        while (*cmd == ' ' || *cmd == '\t') {   // skip spaces
        	cmd++;
        }
        sprintf(tmp, "processTimerCmd() cmd = >%s<", cmd);
        logger("DEBUG", tmp);

        // Do not allow timer commands for Send(string/bytes...)
        if (strncmp(cmd,SEND_STRING, strlen(SEND_STRING)) == 0 ||
            strncmp(cmd,SEND_BYTES,  strlen(SEND_BYTES))  == 0) {
                logger("DEBUG", "processTimerCmd(): incorrect tag in command");                
                return EXIT_NOK;
        }

        float tmout = (float) atof(t1);
        int   times = atoi(t2);

        if (createTimer(id, ID_SET, tmout, times, cmd) == EXIT_NOK) {
        	logger("DEBUG", "processTimerCmd(): timer already exists or input was incorrect");                
                return EXIT_NOK;
        }

        timerCmd* tt = NULL;
        execCmdAndGetResults(ID_SET, cmd, findTimer(id, &tt), NULL);

        return EXIT_OK;
}

void freeTimers()
{
	logger("DEBUG", "freeTimers()");
        timerCmd *tm;

        while (timers != NULL) {
        	tm     = timers;
                timers = timers->next;
        
                free(tm->cmd); tm->cmd = NULL;
                free(tm);
        }
        return;
}

int verifyTimers(int ticksInSec) 
{
	timerCmd *tm   = timers;
        //timerCmd *prev = NULL;

        //int numTimers = 0;
        while (tm != NULL) {
        	
                if (tm->ticks >= (tm->tmout * ticksInSec)) {
                	sprintf(tmp, "verifyTimers(): it is time to execute >%s<", tm->cmd);
                        logger("DEBUG",tmp);

                        execCmdAndGetResults(tm->type, tm->cmd, tm, NULL);
                
                        tm->ticks = 0;
                
                        if (tm->maxTimes > 0) {
                        	tm->times++;
                                if (tm->times >= tm->maxTimes) {    // Cancel this timer 
                                	sprintf(tmp, "verifyTimers(): timer %s self-canceled", tm->id);
                                        logger("DEBUG", tmp);
                                
                                        // Cancel this timer
                                        timerCmd *tmc = tm;
                                        tm = tm->next;

                                        cancelTimer(tmc->id);
                                
                                        continue;
                                }
                        }
                } else {
                	tm->ticks++;
                }
                //prev = tm;
                tm   = tm->next;
                //numTimers++;
        }
	
        return EXIT_OK;
}

static void updateTimerLastData (timerCmd* timer, char *data)
{
	if (timer->lastValue != NULL) {
        	free(timer->lastValue);
                timer->lastValue = NULL;
        }
        timer->lastValue = calloc(strlen(data) + 1, 1);
        strcpy(timer->lastValue,data);
}
